local options = MissionMinder.options;
local L = LibStub("AceLocale-3.0"):GetLocale("MissionMinder");
local DEBUG = MissionMinder.debug;

MissionMinder.planner = {};
local addon = MissionMinder.planner;


local ON_MISSION = GARRISON_FOLLOWER_ON_MISSION; -- "On Mission"
local WORKING = GARRISON_FOLLOWER_WORKING; -- "Working"

local function createInlineTexture(texture)
	return texture and ("|T" .. texture .. ":0|t") or nil;
end

local function canBeSentOnMission(follower)
	return follower and follower.isCollected and ( follower.state == nil or follower.state == ON_MISSION );
end

local function isOnMission(follower)
	return follower and follower.status == ON_MISSION;
end

local function getMissionRewards(mission)
	local rewards = C_Garrison.GetMissionRewardInfo(mission.missionID);

	local t = {};
	for _, reward in pairs(rewards) do
		if(reward.followerXP) then
			local xp = reward.followerXP;
			if ( xp >= 1000 ) then
				xp = xp/1000 .. L["K"];
			end
			tinsert(t, "+" .. xp .. " " .. createInlineTexture(reward.icon));
		elseif (reward.itemID) then
			local itemTexture = select(10, GetItemInfo(reward.itemID));
			if(reward.quantity == 1) then
				tinsert(t, createInlineTexture(itemTexture));
			else
				tinsert(t, reward.quantity .. "x " .. createInlineTexture(itemTexture));
			end
		elseif(reward.currencyID) then
			if (reward.currencyID == 0) then -- gold
				tinsert(t, GetCoinTextureString(reward.quantity));
			else
				tinsert(t, reward.quantity .. " " .. createInlineTexture(reward.icon));
			end
		else
			-- because this is tooltip based, we ignore the value instead of writing the name
		end
	end

	return table.concat(t, ", ");
end

local timeLeft = {};
local inProgressMissions = {};
local function updateFollowerMissionTimeLeft(timeLeft)
	C_Garrison.GetInProgressMissions(inProgressMissions);

	if (inProgressMissions and #inProgressMissions > 0) then
		for _,mission in pairs(inProgressMissions) do
			for _,id in pairs(mission.followers) do
				-- Old values are overwritten, but if a follower is AVAILABLE any previous values will still be present.
				--Thats Okay because we only use this data if the follower is ON_MISSION.
				timeLeft[id] = {
					["missionID"] = mission.missionID,
					["remaining"] = mission.timeLeft;
				};
			end
		end
	end
end

local function levelComparator(left, right)
	return left.level > right.level;
end

local function getMissionTimeleft(follower)
	if isOnMission(follower) then
		-- entry can be nil if follower is on a complete mission that has been viewed, but still needs to be looted.
		local entry = timeLeft[follower.followerID];
		if ( entry == nil or MissionMinder.IsMissionComplete(entry) ) then
			return L["completed"], nil, nil, nil; -- Default yellow
		else
			return string.format("(%s)", timeLeft[follower.followerID].remaining), 0.7, 0.7, 0.7; -- Gray
		end
	elseif (follower.status == WORKING) then
		return L["working"], 1, 0, 0; -- red
	else
		return L["available"], 0, 1, 0; -- Green
	end
end

local function addFollowerCountersToTooltip(self, button)
	local mission = self.info;

	if(mission.inProgress) then -- Do not show tooltip on inProgress missions
		return;
	end


	local followerCounters = C_Garrison.GetBuffedFollowersForMission(mission.missionID);
	local followerTraits = C_Garrison.GetFollowersTraitsForMission(mission.missionID);

	updateFollowerMissionTimeLeft(timeLeft);

	local followers = C_Garrison.GetFollowers(); -- GarrisonMissionFrame.FollowerList.followers
	local numFollowers = followers and #followers or 0;
	table.sort(followers, levelComparator);

	local counterAvailable = false;
	for i1 = 1, numFollowers do
		local follower = followers[i1];

		if ( canBeSentOnMission(follower) ) then
			local counters = followerCounters and followerCounters[follower.followerID] or nil;
			local traits = followerTraits and followerTraits[follower.followerID] or nil;
			local numCounters = counters and #counters or 0;
			local numTraits = traits and #traits or 0;

			if ( numCounters > 0 or numTraits > 0 ) then
				if ( not counterAvailable ) then
					GameTooltip:AddLine(" ");
					GameTooltip:AddLine(L["Counters"]);
					counterAvailable = true;
				end

				local textures = {};
				for i = 1, numCounters do
					local counter = counters[i];
					textures[#textures+1] = createInlineTexture(counter.icon);
				end
				for i = 1, numTraits do
					local trait = traits[i];
					textures[#textures+1] = createInlineTexture(trait.icon);
				end

				local left = string.format("[%d] %s %s", follower.level, follower.name, table.concat(textures, ""));
				local right, r, g, b = getMissionTimeleft(follower);
				GameTooltip:AddDoubleLine(left, right, 1, 1, 1, r, g, b);
			end
		end
	end

	if ( not counterAvailable ) then
		GameTooltip:AddLine(" ");
		GameTooltip:AddLine(L["No counters available"]);
	end

	GameTooltip:Show(); -- resize the tooltip for the new lines
end


local function getFollowerMissionBiasRGB(followerBias)
	if ( followerBias == -1 ) then -- Impossible
		return 1, 0.1, 0.1;
	elseif ( followerBias < 0 ) then -- Hard
		return 1, 0.5, 0.25;
	else -- Normal
		return 1, 1, 1;
	end
end

local function followerHasCounter(abilities, mechanicID, missionType)
	if ( abilities and mechanicID) then
		for a = 1, #abilities do
			local ability = abilities[a];
			for counterID, counterInfo in pairs(ability.counters) do
				if( counterID == mechanicID or counterInfo.name == missionType ) then
					return counterInfo.icon;
				end
			end
		end
	end

	return nil;
end


local expGainTraits = { 80, 81, 29, 236, 98, 25, 99, 30 };


local cache = {
	["isDirty"] = true,
	["abilities"] = {},
	["traits"] = {}
};

local function dirtyMissionCounterCache()
	cache.isDirty = true;
end

local function getFollowerCountersForMisison(followerID, missionID)
	if( cache.isDirty or not cache[missionID] ) then
		cache.abilities[missionID] = C_Garrison.GetBuffedFollowersForMission(missionID);
		cache.traits[missionID] = C_Garrison.GetFollowersTraitsForMission(missionID);
	end

	local abilities = cache.abilities[missionID];
	local traits = cache.traits[missionID];

	local counters = {};
	if ( abilities and abilities[followerID] ) then
		for i = 1, #abilities[followerID] do
			counters[#counters + 1] = createInlineTexture(abilities[followerID][i].icon);
		end
	end

	if ( traits and traits[followerID] ) then
		for i = 1, #traits[followerID] do
			counters[#counters + 1] = createInlineTexture(traits[followerID][i].icon);
		end
	end

	return counters;
end

local function createMissionCountersTooltip(self)
	local follower = C_Garrison.GetFollowerInfo(self.id);
	if ( not follower.isCollected ) then
		return;
	end

	GameTooltip:SetOwner(self, "ANCHOR_NONE");
	GameTooltip:SetPoint("TOPRIGHT", self, "TOPLEFT", -10);

	local r, g, b = GetItemQualityColor(follower.quality);
	GameTooltip:SetText(follower.name, r, g, b);
	GameTooltip:AddLine(" ");

	local availableMissions = GarrisonMissionFrame.MissionTab.MissionList.availableMissions;
	C_Garrison.GetAvailableMissions(availableMissions);
	table.sort(availableMissions, levelComparator);

	local atleastOneMission = false;
	for _,mission in pairs(availableMissions) do
		local bias = C_Garrison.GetFollowerBiasForMission(mission.missionID, follower.followerID);
		r, g, b = getFollowerMissionBiasRGB(bias);

		if ( bias ~= -1 ) then -- don't show impossible (RED) bias missions
			local counters = getFollowerCountersForMisison(follower.followerID, mission.missionID);
			if ( #counters > 0 ) then
				atleastOneMission = true;

				local rewards = getMissionRewards(mission);
				local left = string.format("[%d] %s (%s)", mission.level, mission.name, rewards);
				local right = table.concat(counters, "");
				GameTooltip:AddDoubleLine(left, right, r, g, b, 1, 1, 1);
			end
		end
	end

	if ( not atleastOneMission ) then
		GameTooltip:AddLine(L["No counters available"]);
	end

	GameTooltip:Show();
	cache.isDirty = false;
end


local function hideMissionCountersTooltip(self)
	GameTooltip:Hide();
end


function addon.initialize()
	local missionButtons = GarrisonMissionFrame.MissionTab.MissionList.listScroll.buttons;
	for i = 1, #missionButtons do
		missionButtons[i]:HookScript("OnEnter", addFollowerCountersToTooltip);
	end

	local followerButtons = GarrisonMissionFrame.FollowerList.listScroll.buttons;
	for i = 1, #followerButtons do
		followerButtons[i]:SetScript("OnEnter", createMissionCountersTooltip);
		followerButtons[i]:SetScript("OnLeave", hideMissionCountersTooltip);
	end
end


local events = MissionMinder.events;
events:RegisterEvent("GARRISON_FOLLOWER_LIST_UPDATE", dirtyMissionCounterCache);
events:RegisterEvent("GARRISON_FOLLOWER_XP_CHANGED", dirtyMissionCounterCache);
events:RegisterEvent("GARRISON_FOLLOWER_REMOVED", dirtyMissionCounterCache);